/*
 * SdioDrv.c
 *
 * Copyright (C) 2009 Texas Instruments, Inc. - http://www.ti.com/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation version 2.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any
 * kind, whether express or implied; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

/** \file   SdioDrv.c
 *  \brief  The OMAP3430 Linux SDIO driver (platform and OS dependent)
 *
 * The lower SDIO driver (BSP) for OMAP3430 on Linux OS.
 * Provides all SDIO commands and read/write operation methods.
 *
 *  \see    SdioDrv.h
 */
#include <linux/kernel.h>
#include <linux/slab.h>
#include <mach/io.h>
#include <linux/types.h>
#include <linux/dma-mapping.h>
#include <linux/interrupt.h>
#include <linux/errno.h>
#include <plat/dma.h>
#include <asm/io.h>
#include <linux/clk.h>
#include <plat/clock.h>
#include <plat/resource.h>
#include <linux/platform_device.h>
#include <linux/mmc/core.h>
#include <linux/mmc/card.h>
#include <linux/mmc/host.h>
#include <linux/mmc/sdio_ids.h>
#include <linux/mmc/sdio_func.h>
#include <linux/delay.h>
#include <linux/mmc/mmc.h>
#include <linux/mmc/sdio.h>

#include "SdioDrvDbg.h"
#include "SdioDrv.h"


#define TIWLAN_MMC_MAX_DMA          (8192)

#define SDIO_DRIVER_NAME 			"TIWLAN_SDIO" 
#define SDIO_TOTAL_FUNCS			(2)
#define SDIO_WLAN_FUNC				(2)
#define SDIO_BT_FUNC				(1)
#define SDIO_CTRL_FUNC				(0)

typedef struct OMAP3430_sdiodrv
{
	struct clk    *fclk, *iclk, *dbclk;
	int           dma_tx_channel;
	int           dma_rx_channel;
	void          (*BusTxnCB)(void* BusTxnHandle, int status);
	void*         BusTxnHandle;
	unsigned int  uBlkSize;
	unsigned int  uBlkSizeShift;
	char          *dma_buffer;
	void          *async_buffer;
	unsigned int  async_length;
	int           async_status;
	int (*wlanDrvIf_pm_resume)(void);
	int (*wlanDrvIf_pm_suspend)(void);
	struct device *dev;
	dma_addr_t dma_read_addr;
	size_t dma_read_size;
	dma_addr_t dma_write_addr;
	size_t dma_write_size;
} OMAP3430_sdiodrv_t;



int g_sdio_debug_level = SDIO_DEBUGLEVEL_ERR;


static OMAP3430_sdiodrv_t g_drv;
static struct sdio_func sdio_ctrl_func;
static struct sdio_func *tiwlan_func[1 + SDIO_TOTAL_FUNCS];

 

int sdioDrv_ConnectBus (void *       fCbFunc, 
                        void *       hCbArg, 
                        unsigned int uBlkSizeShift, 
                        unsigned int uSdioThreadPriority)
                       
{
    g_drv.BusTxnCB      = fCbFunc;
    g_drv.BusTxnHandle  = hCbArg;
    g_drv.uBlkSizeShift = uBlkSizeShift;
    g_drv.uBlkSize      = 1 << uBlkSizeShift;

    return 0;
}

int sdioDrv_DisconnectBus (void)
{
    return 0;
}

/*p.3609 cmd flow */
int sdioDrv_ExecuteCmd (unsigned int uCmd, 
                        unsigned int uArg, 
                        unsigned int uRespType, 
                        void *       pResponse, 
                        unsigned int uLen)
{
	struct mmc_command cmd;
	int err;

	PDEBUG("%s: cmd %d\n", __func__, uCmd);

	BUG_ON(!tiwlan_func[SDIO_WLAN_FUNC]->card->host);

	memset(&cmd, 0, sizeof(struct mmc_command));

	cmd.opcode = uCmd;
	cmd.arg = uArg;
	cmd.flags = uRespType;

	err = mmc_wait_for_cmd(tiwlan_func[SDIO_WLAN_FUNC]->card->host, &cmd, 3);
	if (err) {
		printk(KERN_ERR "%s: SDIO Command error, err = %d\n", __func__, err);
	    return -1;
	}

	if (uLen > 0 && uLen <= 4) {
		memcpy(pResponse, (char *)&cmd.resp[0], uLen);
		printk(KERN_DEBUG "%s: response = 0x%x\n", __func__, *(int *)pResponse);
	}

    return 0;
}

static int generic_read_bytes(unsigned int uFunc, unsigned int uHwAddr,
								unsigned char *pData, unsigned int uLen,
								unsigned int bIncAddr, unsigned int bMore)
{
	unsigned int i;
	int ret;

	PDEBUG("%s: uFunc %d uHwAddr %d pData %x uLen %d bIncAddr %d\n", __func__, uFunc, uHwAddr, (unsigned int)pData, uLen, bIncAddr);

	BUG_ON(uFunc != SDIO_CTRL_FUNC && uFunc != SDIO_WLAN_FUNC);
	
    for (i = 0; i < uLen; i++) {
		if (uFunc == 0)
			*pData = sdio_f0_readb(tiwlan_func[uFunc], uHwAddr, &ret);
		else
			*pData = sdio_readb(tiwlan_func[uFunc], uHwAddr, &ret);

		if (0 != ret) {
			printk(KERN_ERR "%s: function %d sdio error: %d\n", __func__, uFunc, ret);
            return -1;
        }

        pData++;
		if (bIncAddr)
			uHwAddr++;
    }

    return 0;
}

static int generic_write_bytes(unsigned int uFunc, unsigned int uHwAddr,
								unsigned char *pData, unsigned int uLen,
								unsigned int bIncAddr, unsigned int bMore)
{
	unsigned int i;
	int ret;

	PDEBUG("%s: uFunc %d uHwAddr %d pData %x uLen %d\n", __func__, uFunc, uHwAddr, (unsigned int) pData, uLen);

	BUG_ON(uFunc != SDIO_CTRL_FUNC && uFunc != SDIO_WLAN_FUNC);

    for (i = 0; i < uLen; i++) {
		if (uFunc == 0)
			sdio_f0_writeb(tiwlan_func[uFunc], *pData, uHwAddr, &ret);
		else
			sdio_writeb(tiwlan_func[uFunc], *pData, uHwAddr, &ret);

		if (0 != ret) {
			printk(KERN_ERR "%s: function %d sdio error: %d\n", __func__, uFunc, ret);
            return -1;
        }

        pData++;
		if (bIncAddr)
			uHwAddr++;
	}

    return 0;
}

/*--------------------------------------------------------------------------------------*/

int sdioDrv_ReadSync (unsigned int uFunc, 
                      unsigned int uHwAddr, 
                      void *       pData, 
                      unsigned int uLen,
                      unsigned int bIncAddr,
                      unsigned int bMore)
{
	int ret;

	PDEBUG("%s: uFunc %d uHwAddr %d pData %x uLen %d bIncAddr %d\n", __func__, uFunc, uHwAddr, (unsigned int)pData, uLen, bIncAddr);

	/* If request is either for sdio function 0 or not a multiple of 4 (OMAP DMA limit)
	    then we have to use CMD 52's */

	if(!tiwlan_func[uFunc]){
		printk("tiwlan_sdio_probe not yet called\n");
		return -1;
	}	

        sdio_claim_host(tiwlan_func[SDIO_WLAN_FUNC]);
	if (uFunc == SDIO_CTRL_FUNC || uLen % 4 != 0)
		ret = generic_read_bytes(uFunc, uHwAddr, pData, uLen, bIncAddr, bMore);
	else
	if (bIncAddr)
		ret = sdio_memcpy_fromio(tiwlan_func[uFunc], pData, uHwAddr, uLen);
	else
		ret = sdio_readsb(tiwlan_func[uFunc], pData, uHwAddr, uLen);
        sdio_release_host(tiwlan_func[SDIO_WLAN_FUNC]);

	if (ret) {
		printk(KERN_ERR "%s: sdio error: %d\n", __func__, ret);
		return -1;
	}

	return 0;
}

/*--------------------------------------------------------------------------------------*/
int sdioDrv_ReadAsync (unsigned int uFunc, 
                       unsigned int uHwAddr, 
                       void *       pData, 
                       unsigned int uLen, 
                       unsigned int bBlkMode,
                       unsigned int bIncAddr,
                       unsigned int bMore)
{
	PERR("%s not yet supported!\n", __func__);

	return -1;

}


/*--------------------------------------------------------------------------------------*/

int sdioDrv_WriteSync (unsigned int uFunc, 
                       unsigned int uHwAddr, 
                       void *       pData, 
                       unsigned int uLen,
                       unsigned int bIncAddr,
                       unsigned int bMore)
{
	int ret;

	PDEBUG("%s: uFunc %d uHwAddr %d pData %x uLen %d bIncAddr %d\n", __func__, uFunc, uHwAddr, (unsigned int)pData, uLen, bIncAddr);

	/* If request is either for sdio function 0 or not a multiple of 4 (OMAP DMA limit)
	    then we have to use CMD 52's */


	if(!tiwlan_func[uFunc]){ 
		printk("tiwlan_sdio_probe not yet called\n");
		return -1;
	}
        sdio_claim_host(tiwlan_func[SDIO_WLAN_FUNC]);
	if (uFunc == SDIO_CTRL_FUNC || uLen % 4 != 0)
		ret = generic_write_bytes(uFunc, uHwAddr, pData, uLen, bIncAddr, bMore);
	else
	if (bIncAddr)
		ret = sdio_memcpy_toio(tiwlan_func[uFunc], uHwAddr, pData, uLen);
	else
		ret = sdio_writesb(tiwlan_func[uFunc], uHwAddr, pData, uLen);

     
        sdio_release_host(tiwlan_func[SDIO_WLAN_FUNC]);
	if (ret) {
		printk(KERN_ERR "%s: sdio error: %d\n", __func__, ret);
		return -1;
	}

	return 0;
}

/*--------------------------------------------------------------------------------------*/
int sdioDrv_WriteAsync (unsigned int uFunc, 
                        unsigned int uHwAddr, 
                        void *       pData, 
                        unsigned int uLen, 
                        unsigned int bBlkMode,
                        unsigned int bIncAddr,
                        unsigned int bMore)
{
	PERR("%s not yet supported!\n", __func__);

	return -1;
}

/*--------------------------------------------------------------------------------------*/

int sdioDrv_ReadSyncBytes (unsigned int  uFunc, 
                           unsigned int  uHwAddr, 
                           unsigned char *pData, 
                           unsigned int  uLen, 
                           unsigned int  bMore)
{
	int err = 0;
	PDEBUG("%s: uFunc %d uHwAddr %d pData %x uLen %d\n", __func__, uFunc, uHwAddr, (unsigned int)pData, uLen);

	if(!tiwlan_func[uFunc]){
		printk("tiwlan_sdio_probe not yet called\n");
		return -1;
	}
        sdio_claim_host(tiwlan_func[SDIO_WLAN_FUNC]);
	err = generic_read_bytes(uFunc, uHwAddr, pData, uLen, 1, bMore);
        sdio_release_host(tiwlan_func[SDIO_WLAN_FUNC]);

	return err;
}

/*--------------------------------------------------------------------------------------*/

int sdioDrv_WriteSyncBytes (unsigned int  uFunc, 
                            unsigned int  uHwAddr, 
                            unsigned char *pData, 
                            unsigned int  uLen, 
                            unsigned int  bMore)
{
	int err = 0;
	PDEBUG("%s: uFunc %d uHwAddr %d pData %x uLen %d\n", __func__, uFunc, uHwAddr, (unsigned int) pData, uLen);
	
	if(!tiwlan_func[uFunc]){
		printk("tiwlan_sdio_probe not yet called\n");
		return -1;
	}

        sdio_claim_host(tiwlan_func[SDIO_WLAN_FUNC]);
        err = generic_write_bytes(uFunc, uHwAddr, pData, uLen, 1, bMore);
        sdio_release_host(tiwlan_func[SDIO_WLAN_FUNC]);

	return err;
}

static void tiwlan_sdio_irq(struct sdio_func *func)
{
    PDEBUG("%s:\n", __func__);
}

int sdioDrv_DisableFunction(unsigned int uFunc)
{
    int err=0;	
    PDEBUG("%s: func %d\n", __func__, uFunc);

	/* currently only wlan sdio function is supported */
	BUG_ON(uFunc != SDIO_WLAN_FUNC);
	BUG_ON(tiwlan_func[uFunc] == NULL);

        sdio_claim_host(tiwlan_func[SDIO_WLAN_FUNC]);
        err = sdio_disable_func(tiwlan_func[SDIO_WLAN_FUNC]);
        sdio_release_host(tiwlan_func[SDIO_WLAN_FUNC]);

	return err;
}

int sdioDrv_EnableInterrupt(unsigned int uFunc)
{
    PDEBUG("%s: func %d\n", __func__, uFunc);

	/* currently only wlan sdio function is supported */
	BUG_ON(uFunc != SDIO_WLAN_FUNC);
	BUG_ON(tiwlan_func[uFunc] == NULL);

	return sdio_claim_irq(tiwlan_func[uFunc], tiwlan_sdio_irq);
}

int sdioDrv_DisableInterrupt(unsigned int uFunc)
{
    PDEBUG("%s: func %d\n", __func__, uFunc);

	/* currently only wlan sdio function is supported */
	BUG_ON(uFunc != SDIO_WLAN_FUNC);
	BUG_ON(tiwlan_func[uFunc] == NULL);

	return sdio_release_irq(tiwlan_func[uFunc]);
}

static int tiwlan_sdio_probe(struct sdio_func *func, const struct sdio_device_id *id)
{
    int err=0;	
    PDEBUG("TIWLAN: probed with vendor 0x%x, device 0x%x, class 0x%x\n",
           func->vendor, func->device, func->class);
    printk("tiwlan_sdio_probe called \n");

    if (func->vendor != SDIO_VENDOR_ID_TI ||
		func->device != SDIO_DEVICE_ID_TI_WL12xx ||
		func->class != SDIO_CLASS_WLAN)
        return -ENODEV;

    printk(KERN_INFO "TIWLAN: Found TI/WLAN SDIO controller (vendor 0x%x, device 0x%x, class 0x%x)\n",
           func->vendor, func->device, func->class);

	tiwlan_func[SDIO_WLAN_FUNC] = func;
	tiwlan_func[SDIO_CTRL_FUNC] = func;
        sdio_claim_host(tiwlan_func[SDIO_WLAN_FUNC]);
        err = sdio_enable_func(tiwlan_func[SDIO_WLAN_FUNC]); 
	err = sdio_set_block_size(tiwlan_func[SDIO_WLAN_FUNC], 512);
	sdio_release_host(tiwlan_func[SDIO_WLAN_FUNC]);

    return 0;
}

static void tiwlan_sdio_remove(struct sdio_func *func)
{
    int err=0;
    PDEBUG("%s\n", __func__);
    printk("%s\n", __func__);

    /*sdio_release_host(tiwlan_func[SDIO_WLAN_FUNC]);*/
}

static const struct sdio_device_id tiwl12xx_devices[] = {
       {.class = SDIO_CLASS_WLAN,
	   	.vendor = SDIO_VENDOR_ID_TI,
	   	.device = SDIO_DEVICE_ID_TI_WL12xx},
       {}
};
MODULE_DEVICE_TABLE(sdio, tiwl12xx_devices);

static struct sdio_driver tiwlan_sdio_drv = {
    .probe          = tiwlan_sdio_probe,
    .remove         = tiwlan_sdio_remove,
    .name           = "sdio_tiwlan",
    .id_table       = tiwl12xx_devices,
};

static int __init sdioDrv_init(void)
{
	int ret;

	PDEBUG("%s: Debug mode\n", __func__);
	printk("%s: Debug mode\n", __func__);

	memset(&g_drv, 0, sizeof(g_drv));

	g_drv.dma_buffer = kzalloc(TIWLAN_MMC_MAX_DMA, GFP_KERNEL|GFP_DMA);
	if (g_drv.dma_buffer == NULL) {
	  ret = -ENOMEM;
	  goto out;
	}

    ret = sdio_register_driver(&tiwlan_sdio_drv);
    if (ret < 0) {
        printk(KERN_ERR "sdioDrv_init: sdio register failed: %d\n", ret);
		kfree(g_drv.dma_buffer);
		goto out;
    }

	printk(KERN_INFO "TI WiLink 1271 SDIO: Driver loaded\n");

out:
	return ret;
}

static void __exit sdioDrv_exit(void)
{
	sdio_unregister_driver(&tiwlan_sdio_drv);

	kzfree(g_drv.dma_buffer);

	printk(KERN_INFO "TI WiLink 1271 SDIO Driver unloaded\n");
}

module_init(sdioDrv_init);
module_exit(sdioDrv_exit);
module_param(g_sdio_debug_level, int, SDIO_DEBUGLEVEL_ERR);
MODULE_PARM_DESC(g_sdio_debug_level, "TIWLAN SDIO debug level");
EXPORT_SYMBOL(g_sdio_debug_level);
EXPORT_SYMBOL(sdioDrv_ConnectBus);
EXPORT_SYMBOL(sdioDrv_DisconnectBus);
EXPORT_SYMBOL(sdioDrv_ExecuteCmd);
EXPORT_SYMBOL(sdioDrv_ReadSync);
EXPORT_SYMBOL(sdioDrv_WriteSync);
EXPORT_SYMBOL(sdioDrv_ReadAsync);
EXPORT_SYMBOL(sdioDrv_WriteAsync);
EXPORT_SYMBOL(sdioDrv_ReadSyncBytes);
EXPORT_SYMBOL(sdioDrv_WriteSyncBytes);
EXPORT_SYMBOL(sdioDrv_DisableFunction);
EXPORT_SYMBOL(sdioDrv_EnableInterrupt);
EXPORT_SYMBOL(sdioDrv_DisableInterrupt);
MODULE_DESCRIPTION("TI WLAN 1271 SDIO interface");
MODULE_LICENSE("GPL");
MODULE_ALIAS(SDIO_DRIVER_NAME);
MODULE_AUTHOR("Texas Instruments Inc");

